/*
Copyright (C) 2001-2002 pat@aftermoon.net

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// g_wsrv.c -- WorldServer Remote Procedure Call for game dll (pat@aftermoon.net)

#include "../g_local.h"
#include "wsrv.h"
#include "wsrv_rpc.h"

//-----------------------------------------------------------------------------
// CONST
//-----------------------------------------------------------------------------

// !!! if this is changed, it must be changed in qcommon.h too !!!
#define PORT_SERVER	    27910

#define WSRV_SEND_DATA_PERIOD 30000	// 30 seconds

//-----------------------------------------------------------------------------
// VARIABLES
//-----------------------------------------------------------------------------

static int l_WSrv_time = 0 ;

// Variable for debug log file
static FILE    *l_RPCLOG_file ;
static char     l_RPCLOG_Text[4096] ;
static int      l_RPCLOG_Len = 0 ;
static int      l_RPCLOG_Count = 0 ;

// Internal data for WorldServer link management
wsrv_data_t g_WsrvData;

//-----------------------------------------------------------------------------
// Return a pointer on a client struct with a given name (or NULL if no match)
//-----------------------------------------------------------------------------
gclient_t *GetClientByName(char *szName)
{
	int	i;

	for ( i=0 ; i<g_maxclients.integer ; i++ )
	{
		if ( !level.clients[i].pers.connected )
			continue;
		
		if ( stricmp(level.clients[i].pers.netname, szName)==0 )
		{
			return &level.clients[i] ;
		}
	}

	return NULL;
}

//-----------------------------------------------------------------------------
// Flush the message loging buffer
//-----------------------------------------------------------------------------
void RPC_FlushLog(char *pszLog, int iSizeMax)
{
    l_RPCLOG_Count++ ;
    strncpy(pszLog, l_RPCLOG_Text, iSizeMax-1) ;
    pszLog[iSizeMax-1] = '\0' ;

    l_RPCLOG_Len = 0 ;
    l_RPCLOG_Text[0] = '\0' ;
}

//-----------------------------------------------------------------------------
// Message log for the WorldServer
// The last message is always lost for the worldserver (but not for the local 
// file) because each message send the log of the previous . At each dll reload 
// we lost a log message, it's not a bug, it's a feature ;-)
//-----------------------------------------------------------------------------
void RPC_Log(char *fmt, ...)
{
	va_list		argptr;
    char        *pszTmp ;
	char		msg[16384];
    int         iLen ;

	va_start (argptr,fmt);
	vsprintf (msg,fmt,argptr);
    va_end (argptr);

	RPC_LogFile(msg);

    pszTmp = strchr(msg, '&') ;
    while ( pszTmp )
    {
        *pszTmp = '#' ;
        pszTmp = strchr(pszTmp, '&') ;
    }

    // Store message for sending it to the WorldServer
    iLen = strlen(msg) ;
    if ( iLen>128 )
        iLen = 128 ;
    if ( (sizeof(l_RPCLOG_Text) - l_RPCLOG_Len) > (iLen+7) )
    {
        strncpy(l_RPCLOG_Text+l_RPCLOG_Len, msg, iLen) ;
        l_RPCLOG_Text[l_RPCLOG_Len+iLen] = '\0' ;
        l_RPCLOG_Len += iLen ;
    }
    else if ( (sizeof(l_RPCLOG_Text) - l_RPCLOG_Len) > 7 )
    {
        strcpy(l_RPCLOG_Text+l_RPCLOG_Len, "<full>") ;
        l_RPCLOG_Len += 6 ;
    }
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void RPC_LogFile(char *msg)
{
    if (!l_RPCLOG_file)
		l_RPCLOG_file = fopen ("wsrv.log", "a");

    if (l_RPCLOG_file)
    {
		fprintf (l_RPCLOG_file, msg);
		fflush (l_RPCLOG_file);          // force it to save every time
    }
}

//-----------------------------------------------------------------------------
// Light XML parser for simple tag
//-----------------------------------------------------------------------------
static qboolean XML_Parser(const char *pszData, char *pszTag, char *pszVar, int iSize)
{
    char    szBegin[64], szEnd[64] ;
    char    *pszBegin, *pszEnd ;

    // Default value is empty string
    *pszVar = '\0' ;

    Com_sprintf( szBegin, sizeof(szBegin), "<%s>", pszTag );
    pszBegin = strstr(pszData, szBegin) ;
    if ( pszBegin )
    {
        pszBegin += strlen(szBegin) ;

        Com_sprintf( szEnd, sizeof(szEnd), "</%s>", pszTag );
        pszEnd = strstr(pszBegin, szEnd) ;
        if ( pszEnd )
        {
            if ( iSize >= (pszEnd - pszBegin) )
            {
                iSize = pszEnd - pszBegin ;
                memcpy(pszVar, pszBegin, iSize) ;
                pszVar[iSize] = '\0';
                return qtrue;
            }
        }
    }
    return qfalse;
}

//-----------------------------------------------------------------------------
// Treat a RPC error
//-----------------------------------------------------------------------------
void RPC_Error(volatile rpc_data_t *pRpcData, char *pszErrMsg, int iSizeErrMsg)
{
    char        szStatus[16];
    char        szErrCode[16];
    char        szErrMsg[64];

    if (    XML_Parser(pRpcData->pszResult, "status",  szStatus,  sizeof(szStatus))
         && strcmp(szStatus, "ERR")==0
         && XML_Parser(pRpcData->pszResult, "errcode", szErrCode, sizeof(szErrCode))
         && XML_Parser(pRpcData->pszResult, "errmsg",  szErrMsg,  sizeof(szErrMsg)) )
    {
        Com_sprintf( pszErrMsg, iSizeErrMsg, "%s\0%s", szErrMsg, szErrCode );
    }
    else
        Com_sprintf( pszErrMsg, iSizeErrMsg, "Invalid WorldServer reply" );

    Com_Printf ("### RPC_Error : %s\n", pszErrMsg);
}

//-----------------------------------------------------------------------------
// Manage common tag that can be received with every worldserver answer
//
//	<cvar_set> string </cvar_set>
//
//	<cvar_forceset> string </cvar_forceset>
//
//	<wsrv_internal>
//		<startmap> string </startmap>
//		<flags> string </flags>
//	</wsrv_internal>
//
//	<bprintf> message </bprintf>
//
//	<serverlist>
//		<adr1> ip:port </adr1>
//		<adr2>
//		:
//		<adr8>
//	</serverlist>
//
//	<client_set>
//		<client1>
//			<name> string </name>
//			<health> integer </health>
//			<inventory> string <inventory>
//			<spectator> integer </spectator>
//		</client1>
//		<client2>
//		:
//		<client8>
//	</client_set>
//
//-----------------------------------------------------------------------------
void WSrv_CommonReplyTag_OK(char *pszResult)
{
    char    szTmp[512], szTmp2[512];
    char    *pszCvar, *pszVal ;
    char    szData[1024], szSubData[1024];
    int     i;
//    char    szName[16];
//    char    *pszItem, *pszNb ;
//    int     iNb,iItem;
//	gclient_t *pClient;


    // A CVAR_SET is requested by the worldserver
    if ( XML_Parser(pszResult, "cvar_set",  szTmp,  sizeof(szTmp)) )
    {
        strtok(szTmp, "<") ;
        while ( pszCvar = strtok(NULL, ">") )
        {
            pszVal = strtok(NULL, "<");
            trap_Cvar_Set( pszCvar, pszVal );
            strtok(NULL, "<");      // next Cvar
        }
    }

    // A CVAR_FORCESET is requested by the worldserver
    if ( XML_Parser(pszResult, "cvar_forceset",  szTmp,  sizeof(szTmp)) )
    {
        strtok(szTmp, "<") ;
        while ( pszCvar = strtok(NULL, ">") )
        {
            pszVal = strtok(NULL, "<");
            trap_Cvar_Set( pszCvar, pszVal );	// FIXME : forceset
            strtok(NULL, "<");      // next Cvar
        }
    }

    // An internal WorldServer data is requested by the worldserver
    if ( XML_Parser(pszResult, "wsrv_internal",  szTmp,  sizeof(szTmp)) )
	{
	    XML_Parser(pszResult, "startmap", g_WsrvData.szStartMap, sizeof(g_WsrvData.szStartMap));
	    if ( XML_Parser(pszResult, "flags",  szTmp2,  sizeof(szTmp2)) )
			g_WsrvData.dwFlags = (DWORD)atol(szTmp2);
	}

    // Receive a broadcast message from the WorldServer
    if ( XML_Parser(pszResult, "bprintf", szData, sizeof(szData)) )
    {
        //gi.bprintf(PRINT_MEDIUM, va("%s\n", szData)) ;
    }

    // Receive the server list from the WorldServer
    if ( XML_Parser(pszResult, "serverlist", szData, sizeof(szData)) )
    {
        for ( i=0 ; i<8 ; i++ ) 
        {
            if ( XML_Parser(pszResult, va("adr%d", i), szSubData, sizeof(szSubData)) )
                trap_SendServerCommand(-1, va("adr%d %s\n", i, szSubData));
            else
                trap_SendServerCommand(-1, va("adr%d \"\"\n", i));
        }
    }

    // Receive a playergetinfo answer from the WorldServer
    if ( XML_Parser(pszResult, "client_set", szData, sizeof(szData)) )
    {
/*$$
		for ( i=0 ; i<g_maxclients.integer ; i++ )
		{
			if (    XML_Parser(szData, va("client%d", i), szSubData, sizeof(szSubData))
				 && XML_Parser(szSubData, "name", szName, sizeof(szName))
				 && (pClient = GetClientByName(szName)) != NULL )
			{
				if ( XML_Parser(szSubData, "health", szTmp,    sizeof(szTmp)) )
					pClient->pers.health = atoi(szTmp);

				if ( XML_Parser(szSubData, "inventory", szTmp, sizeof(szTmp)) )
				{
					// Reset inventory, before filling it
					memset(pClient->pers.inventory, 0, sizeof(level.clients[0].pers.inventory));

					pszItem = strtok(szTmp, ":") ;
					while ( pszItem )
					{
						pszNb = strtok(NULL, ",") ;
						if ( pszNb )
						{
							iNb   = atoi(pszNb) ;
							iItem = atoi(pszItem);
            
							pClient->pers.inventory[iItem] = iNb ;
						}
						pszItem = strtok(NULL, ":") ;
					}
				}

				if ( XML_Parser(szSubData, "spectator", szTmp, sizeof(szTmp)) )
					pClient->pers.spectator = atoi(szTmp) != 0;
			}
		}
*/
	}
}

//-----------------------------------------------------------------------------
// Called by the standard Killed function for reporting each kill to the WorldServer
//-----------------------------------------------------------------------------
#define MOD_SELF_FIRE   0x4000000
//-----------------------------------------------------------------------------
void WSrv_Killed(gentity_t *targ, gentity_t *inflictor, gentity_t *attacker, int meansOfDeath)
{
    char    *pszAttacker, *pszCommand, *pszTarg ;
	char	szData[128];
	char	str[MAX_INFO_VALUE];

    // Force immediat send of data, because a player death is a priority data
    l_WSrv_time = -WSRV_SEND_DATA_PERIOD ;

	if ( targ && targ->client )
		pszTarg = targ->client->pers.netname ;
	else if ( targ )
		pszTarg = targ->classname ;
	else
		pszTarg = "???";

	if (attacker && attacker->client)
    {
        pszCommand  = "pk" ;
        pszAttacker = attacker->client->pers.netname ;
    }
    else
    {
        pszCommand  = "kb" ; 
        if (attacker)
            pszAttacker = attacker->classname ;
        else
            pszAttacker = "" ;
    }

	trap_Cvar_VariableStringBuffer( "mapname", str, sizeof(str) );
	Com_sprintf(szData, sizeof(szData), "%s,%d,%d,%d",
		str,
		(int)targ->s.origin[0], 
		(int)targ->s.origin[1], 
		(int)targ->s.origin[2]);
    
    WSrv_GameLoopAddData(va("%s,%s,%s,%X,%s/",pszAttacker, pszCommand, pszTarg,meansOfDeath,szData));
}

//-----------------------------------------------------------------------------
// Inform the WorldServer that a player quit the server
//-----------------------------------------------------------------------------
void WSrv_ClientDisconnect(const char *pszName)
{
    WSrv_GameLoopAddData(va("%s,quit/", pszName));
}

//-----------------------------------------------------------------------------
// Inform the WorldServer that a player join the server
//-----------------------------------------------------------------------------
void WSrv_ClientConnect(gentity_t *ent)
{
    WSrv_AddCmd(va("GetInventory,%s", ent->client->pers.netname));


    if ( !ent->client->pers.localClient )    // Inform GameLoop only if it's not the server
        WSrv_GameLoopAddData(va("%s,join/", ent->client->pers.netname));
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
static char l_szGameLoopData[4096] = "" ;
//-----------------------------------------------------------------------------
void WSrv_GameLoopAddData(const char *szData)
{
    int iLen = strlen(l_szGameLoopData) ;

	if ( (iLen+strlen(szData)) > sizeof(l_szGameLoopData) )
		RPC_Log("ERROR\tGameLoopData: overflow of %i in %i (%s)", 
			iLen+strlen(szData), sizeof(l_szGameLoopData), szData);
	else
	    Com_sprintf ( l_szGameLoopData+iLen, sizeof(l_szGameLoopData)-iLen, "%s", szData ) ;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void WSrv_GameLoop(volatile rpc_data_t *pRpcData)
{
	gentity_t		*cl_ent;
	static char	buffer[4096];
    int     i,iCount;

	Com_Printf("### WSRV call GameLoop\n");

	iCount = 0;
    for ( i=0 ; i<g_maxclients.integer ; i++ )
    {
        if ( level.clients[i].pers.connected )
		{
			cl_ent = &g_entities[i];
            WSrv_GameLoopAddData(va("playername,%s,ping,%d,health,%d,x,%d,y,%d,z,%d/", 
				cl_ent->client->pers.netname, 
				cl_ent->client->ps.ping,
				cl_ent->health,
				(int)cl_ent->s.origin[0], 
				(int)cl_ent->s.origin[1], 
				(int)cl_ent->s.origin[2]) );
			iCount++;
		}
    }
	if ( !iCount )
		RPC_Log("WARNING\tWSrv_GameLoop\tiCount=0\tg_maxclients.integer=%0d\n", g_maxclients.integer);

	Com_sprintf ( buffer, sizeof(buffer), "%s?func=GameLoop&id=%s&session=%s&data=%s",
                  g_WsrvData.szUrl, 
                  g_WsrvData.szId, 
                  g_WsrvData.szSession,
                  l_szGameLoopData ) ;

    pRpcData->iData      = WSRV_GAMELOOP ;
    pRpcData->pszRequest = buffer ;
    pRpcData->iFlags     = 0 ;
    HTTP_SendRpc(pRpcData) ;

    // Reset the command buffer
    l_szGameLoopData[0] = '\0' ;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void WSrv_GameLoop_OK(char *pszResult)
{
    WSrv_CommonReplyTag_OK(pszResult);
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
extern int FS_LoadFile (char *path, void **buffer);
extern void FS_FreeFile (void *buffer);
//----------------------------------------------------------------------------
void WSrv_MapGetXmlRules(volatile rpc_data_t *pRpcData)
{
	static char		buffer[320];
    int				i, iLen, iNbPlayer;
    char			mapchecksum[MAX_INFO_VALUE];
	char			mapname[MAX_INFO_VALUE];

	trap_Cvar_VariableStringBuffer("sv_mapChecksum", mapchecksum, sizeof(mapchecksum));
	trap_Cvar_VariableStringBuffer("mapname", mapname, sizeof(mapname));

	Com_sprintf ( buffer, sizeof(buffer), 
                  "%s?func=MapGetXmlRules&id=%s&session=%s"
                  "&mapname=%s&mapchecksum=%s&player_list=",
                  g_WsrvData.szUrl,
                  g_WsrvData.szId, 
                  g_WsrvData.szSession,
                  mapname,
				  mapchecksum ) ;

    iLen = strlen(buffer) ;
    iNbPlayer = 0 ;
    for ( i=0 ; i<g_maxclients.integer ; i++ )
    {
		if ( level.clients[i].pers.connected )
        {
            Com_sprintf ( buffer+iLen, 
                          sizeof(buffer)-iLen, 
                          "%s,", level.clients[i].pers.netname ) ;
            iNbPlayer++ ;
            iLen += strlen(buffer+iLen) ;
        }
    }

    if ( !iNbPlayer )
    {
		char	name[MAX_INFO_VALUE];
		
		trap_Cvar_VariableStringBuffer("name", name, sizeof(name));
        if ( strlen(name) )
            Com_sprintf (buffer+iLen, sizeof(buffer)-iLen, "%s", name) ;
    }
    else
        buffer[iLen-1] = '\0' ;     // Suppress the last ','

    pRpcData->iData      = WSRV_MAPGETXMLRULES ;
    pRpcData->pszRequest = buffer ;
    pRpcData->iFlags     = 0 ;
    HTTP_SendRpc(pRpcData) ;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
qboolean WSrv_MapGetXmlRules_OK(volatile rpc_data_t *pRpcData)
{
	char		szStatus[4];
	qboolean	bResult = qfalse;

    if ( pRpcData->byRpcStatus == RPC_STATUS_OK )
    {
		if (    XML_Parser(pRpcData->pszResult, "status", szStatus, sizeof(szStatus))
			 && strcmp(szStatus, "OK")==0 )
		{
			WSrv_CommonReplyTag_OK(pRpcData->pszResult);
			bResult = qtrue;
		}
	}
	return bResult;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void WSrv_GetInventory(volatile rpc_data_t *pRpcData, char *pszName)
{
	static char	buffer[1024];

	Com_sprintf ( buffer, sizeof(buffer), 
                  "%s?func=PlayerGetInfo&id=%s&session=%s"
                  "&player_info_name=%s&token=get_inventory",
                  g_WsrvData.szUrl, 
                  g_WsrvData.szId, 
                  g_WsrvData.szSession,
                  pszName ) ;

    if ( strlen(buffer) >= sizeof(buffer)-1 )
        Com_Printf("WSrv_GetInventory : buffer too small (%d)\n", sizeof(buffer));

    pRpcData->iData      = WSRV_PLAYERGETINFO ;
    pRpcData->pszRequest = buffer ;
    pRpcData->iFlags     = 0 ;
    HTTP_SendRpc(pRpcData) ;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void WSrv_GetInventory_OK(char *pszResult)
{
	WSrv_CommonReplyTag_OK(pszResult);
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void WSrv_SetInventory(volatile rpc_data_t *pRpcData, char *apszName[], int iNbPlayer)
{
	static char	buffer[4096];
    int         iPlayer, i, iLen ;
//	int			ii;
	char		mapname[MAX_INFO_VALUE];

	trap_Cvar_VariableStringBuffer("mapname", mapname, sizeof(mapname));

	Com_sprintf ( buffer, sizeof(buffer), 
                  "%s?func=PlayersSetInfo&id=%s&session=%s",
                  g_WsrvData.szUrl, 
                  g_WsrvData.szId, 
                  g_WsrvData.szSession ) ;

    iLen = strlen(buffer) ;

    for ( iPlayer=0 ; iPlayer < iNbPlayer ; iPlayer++ )
    {
        // check for a name match
	    for ( i=0 ; i<g_maxclients.integer ; i++ )
	    {
		    if ( !level.clients[i].pers.connected )
			    continue;
		    if ( stricmp(level.clients[i].pers.netname, apszName[iPlayer])==0 )
		    {
                // Save inventory
                Com_sprintf ( buffer+iLen, sizeof(buffer)-iLen,
                              "&player_info_name%d=%s"
                              "&data_map%d=%s"
                              "&data_health%d=%d"
                              "&data_inventory%d=",
                              iPlayer, apszName[iPlayer],
                              iPlayer, mapname,
                              iPlayer, g_entities[i].health,
                              iPlayer ) ;
                iLen += strlen(buffer+iLen) ;
/*$$
	            for ( ii=0 ; ii<game.num_items ; ii++ )
                {
		            if ( level.clients[i].pers.inventory[ii] )
                    {
	                    Com_sprintf ( buffer+iLen, sizeof(buffer)-iLen,
                                      "%d:%d,", ii, level.clients[i].pers.inventory[ii] ) ;
                        
                        iLen += strlen(buffer+iLen) ;
                    }
                }
$$*/
                if ( buffer[iLen-1]==',' )
                    buffer[--iLen] = '\0' ; // Suppress the last comma

                break;
		    }
	    }
        if ( i >= g_maxclients.integer )
            Com_Printf("### WSrv_SetInventory : Player %s not found\n", apszName[iPlayer]);
    }

    if ( iLen >= sizeof(buffer)-1 )
        Com_Printf("WSrv_SetInventory : buffer too small (%d)\n", iLen);

    pRpcData->iData      = WSRV_PLAYERSSETINFO ;
    pRpcData->pszRequest = buffer ;
    pRpcData->iFlags     = 0 ;
    HTTP_SendRpc(pRpcData) ;
}

//-----------------------------------------------------------------------------
//    FIFO
//-----------------------------------------------------------------------------
static char WSrv_FifoCmd[15][40] ;
static int  WSrv_FifoIn  = 0 ;
static int  WSrv_FifoOut = 0 ;

//-----------------------------------------------------------------------------
//    Add a command in the FIFO
//-----------------------------------------------------------------------------
void WSrv_AddCmd(char *pszCmd)
{
    if ( ((WSrv_FifoIn + 1) % SIZE_TAB(WSrv_FifoCmd)) == WSrv_FifoOut )
        return ;

    strncpy(WSrv_FifoCmd[WSrv_FifoIn], pszCmd, sizeof(WSrv_FifoCmd[0]));
    WSrv_FifoCmd[WSrv_FifoIn][sizeof(WSrv_FifoCmd[0])-1] = 0 ;
    WSrv_FifoIn++ ;
    WSrv_FifoIn %= SIZE_TAB(WSrv_FifoCmd) ;
}

//-----------------------------------------------------------------------------
//  Send a command to the WorldServer
//-----------------------------------------------------------------------------
void WSrv_SendCmd(volatile rpc_data_t *pRpcData)
{
    char        *pszFunc ;
    char        *pszName ;

    // Treat a command
    pszFunc = strtok(WSrv_FifoCmd[WSrv_FifoOut], ",");
    WSrv_FifoOut++ ;
    WSrv_FifoOut %= SIZE_TAB(WSrv_FifoCmd) ;

    if ( strcmp(pszFunc,"GetInventory")==0 )
    {
        pszName = strtok(NULL, ",");
        WSrv_GetInventory(pRpcData, pszName) ;
    }
    else if ( strcmp(pszFunc,"SetInventory")==0 )
    {
        int     iNbPlayer = 0 ;
        char    *apszNames[4] ;

        pszName = strtok(NULL, ",");
        apszNames[iNbPlayer++] = pszName ;

        // If multiple SetInventory in the stack : Concatenate them
        while (    WSrv_FifoIn != WSrv_FifoOut
                && iNbPlayer < SIZE_TAB(apszNames)
                && strncmp(WSrv_FifoCmd[WSrv_FifoOut], "SetInventory,", strlen("SetInventory,"))==0 )
        {
            pszFunc = strtok(WSrv_FifoCmd[WSrv_FifoOut], ",");
            pszName = strtok(NULL, ",");
            apszNames[iNbPlayer++] = pszName ;

            WSrv_FifoOut++ ;
            WSrv_FifoOut %= SIZE_TAB(WSrv_FifoCmd) ;
        }

        WSrv_SetInventory(pRpcData, apszNames, iNbPlayer) ;
    }
    else if ( strcmp(pszFunc,"Logout")==0 )
    {
        RPC_Logout(pRpcData) ;
    }
}

//-----------------------------------------------------------------------------
//  Send immediately a command to the WorldServer
//-----------------------------------------------------------------------------
void WSrv_SendCmdImmediate(char *pszCmd)
{
    static char RpcData_buffer[256] ;
    static volatile rpc_data_t  RpcData = { RPC_NONE, 0, NULL, 
        RpcData_buffer, sizeof(RpcData_buffer), RPC_STATUS_READY } ;
    char        *pszFunc ;
    char        *pszName ;

    if ( RpcData.byRpcStatus != RPC_STATUS_IN_PROGRESS )
    {
        pszFunc = strtok(pszCmd, ",");

        if ( strcmp(pszFunc,"SetInventory")==0 )
        {
            pszName = strtok(NULL, ",");
            WSrv_SetInventory(&RpcData, &pszName, 1) ;
        }
    }
    else
	    Com_Printf("### WSRV impossible to send \"%s\"\n", pszCmd);
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
static char l_RpcData_buffer[RPC_RESULT_SIZE] ;
static volatile rpc_data_t  l_RpcData = { RPC_NONE, 0, NULL, 
    l_RpcData_buffer, sizeof(l_RpcData_buffer), RPC_STATUS_READY } ;
//-----------------------------------------------------------------------------
void WSrv_CheckWorldServer(qboolean bFlush)
{
    char    szErrMsg[80] ;
    char    szStatus[4] ;

    // Force immediat send of data if the level has changed or if bFlush
    if ( bFlush || (l_WSrv_time > level.time) )
        l_WSrv_time = -WSRV_SEND_DATA_PERIOD ;

    do
    {
        if ( l_RpcData.byRpcStatus == RPC_STATUS_READY )
        {
            int   iLen = strlen(l_szGameLoopData) ;

            if (    iLen > (sizeof(l_szGameLoopData) * 0.75)
                      || ((l_WSrv_time + WSRV_SEND_DATA_PERIOD) < level.time) )
            {   // Then 
                l_WSrv_time = level.time ;
                WSrv_GameLoop(&l_RpcData) ;
            }
            else if ( WSrv_FifoIn != WSrv_FifoOut )
            {   // Priority to the WorldServer command stack
                WSrv_SendCmd(&l_RpcData) ;
            }
            else
                bFlush = qfalse ;
        }
        else if ( l_RpcData.byRpcStatus == RPC_STATUS_OK )
        {
            if (    XML_Parser(l_RpcData.pszResult, "status",  szStatus,  sizeof(szStatus))
                 && strcmp(szStatus, "OK")==0 )
            {
                if ( l_RpcData.iData == WSRV_PLAYERGETINFO )
                    WSrv_GetInventory_OK(l_RpcData.pszResult) ;
                else if ( l_RpcData.iData == WSRV_GAMELOOP )
                    WSrv_GameLoop_OK(l_RpcData.pszResult) ;
                else if ( l_RpcData.iData == RPC_LOGOUT )
                {
                    // It's not really a Logout, it can be a server change
                }
            }
            else
                RPC_Error(&l_RpcData, szErrMsg, sizeof(szErrMsg));
        
            l_RpcData.byRpcStatus = RPC_STATUS_READY ;
        }
        else if ( l_RpcData.byRpcStatus == RPC_STATUS_ERROR )
        {
            Com_Printf ("### RPC_Error : no reply\n");
            l_RpcData.byRpcStatus = RPC_STATUS_READY ;
            bFlush = qfalse ;
        }
        else  if ( bFlush && (l_RpcData.byRpcStatus == RPC_STATUS_IN_PROGRESS) )
        {
            // If we are looping, we don't need to charge the CPU too much
            RPC_Sleep(50) ;
        }
    }
    while ( bFlush ) ;
}

#ifndef WSRV_VERSION
#define WSRV_VERSION "0.3b"
#endif

#ifndef WSRV_ID_EXE
#define WSRV_ID_EXE "version_" WSRV_VERSION "_" __DATE__
#endif
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void RPC_Login(volatile rpc_data_t *pRpcData, char *szName, char *szPass)
{
	static char	buffer[256] ;
	int			i ;
	char		version[MAX_INFO_VALUE];

	trap_Cvar_VariableStringBuffer( "version", version, sizeof(version) );

    // Verify if the '/' is present, then add it if not (it's a current user mistake)
    i = strlen(g_WsrvData.szUrl) ;
    if ( i && g_WsrvData.szUrl[i-1]!='/' )
    {
        char    szBuffer[80] ;

        i = min(i, sizeof(szBuffer)-2) ;
        strncpy(szBuffer, g_WsrvData.szUrl, i) ;
        szBuffer[i]   = '/' ;
        szBuffer[i+1] = 0 ;
        trap_Cvar_Set("wsrv_url", szBuffer);	// FIXME : forceset
    }

	Com_sprintf ( buffer, sizeof(buffer), "%s?func=PlayerLogin&name=%s&pass=%s&version=%s&wsrv_version=%s&idexe=%s",
                  g_WsrvData.szUrl, 
                  szName,
                  szPass,
				  version,
                  WSRV_VERSION,
                  WSRV_ID_EXE );

    pRpcData->iData = RPC_LOGIN ;
    pRpcData->pszRequest = buffer ;
    pRpcData->iFlags = 0 ;
    HTTP_SendRpc(pRpcData) ;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
qboolean RPC_Login_OK(volatile rpc_data_t *pRpcData, char *pszErrMsg, int iSizeErrMsg)
{
    qboolean    bResult = qfalse;
    char        szStatus[16];

    if (    XML_Parser(pRpcData->pszResult, "status",  szStatus,  sizeof(szStatus))
         && strcmp(szStatus, "OK")==0
         && XML_Parser(pRpcData->pszResult, "id",      g_WsrvData.szId,      sizeof(g_WsrvData.szId))
         && XML_Parser(pRpcData->pszResult, "session", g_WsrvData.szSession, sizeof(g_WsrvData.szSession)) )
    {
		WSrv_CommonReplyTag_OK(pRpcData->pszResult);
        bResult = qtrue;
    }
    else
        RPC_Error(pRpcData, pszErrMsg, iSizeErrMsg);

    return bResult;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void RPC_PlayerCreateGame(volatile rpc_data_t *pRpcData)
{
	static char	buffer[256];

//$$    if ( port->value < 1 )
//$$        trap_Cvar_Set("port", va("%d",PORT_SERVER));	// FIXME : forceset

    Com_sprintf ( buffer, sizeof(buffer), "%s?func=PlayerCreateGame&id=%s&session=%s&port=%s&dedicated=%d",
                  g_WsrvData.szUrl, 
                  g_WsrvData.szId, 
                  g_WsrvData.szSession,
                  0, //$$ port->string, 
                  g_dedicated.integer);

    pRpcData->iData = RPC_PLAYER_CREATEGAME ;
    pRpcData->pszRequest = buffer ;
    pRpcData->iFlags = 0 ;
    HTTP_SendRpc(pRpcData) ;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
qboolean RPC_PlayerCreateGame_OK(volatile rpc_data_t *pRpcData, char *pszMsg, int iSizeMsg)
{
    qboolean    bResult = qfalse;
    char        szStatus[8];

    if (    XML_Parser(pRpcData->pszResult, "status",  szStatus,  sizeof(szStatus))
         && strcmp(szStatus, "OK")==0 )
    {
        WSrv_CommonReplyTag_OK(pRpcData->pszResult);
        bResult = qtrue;
    }
    else
        RPC_Error(pRpcData, pszMsg, iSizeMsg);

    return bResult;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void RPC_Logout(volatile rpc_data_t *pRpcData)
{
	static char	buffer[256];

	Com_sprintf ( buffer, sizeof(buffer), "%s?func=PlayerLogout&id=%s&session=%s",
                  g_WsrvData.szUrl, 
                  g_WsrvData.szId, 
                  g_WsrvData.szSession);

    pRpcData->iData = RPC_LOGOUT ;
    pRpcData->pszRequest = buffer ;
    pRpcData->iFlags = 0 ;
    HTTP_SendRpc(pRpcData) ;
}

typedef struct {
	vmCvar_t	*vmCvar;
	char		*cvarName;
	char		*defaultString;
	int			cvarFlags;
	int			modificationCount;  // for tracking changes
} cvarTable_t;

//-----------------------------------------------------------------------------
// Console command : Login to a WorldServer
// The "ent" parameter could be NULL if it's a dedicated server
// If we are already connected to the worldserver, this command
// jump to the StartMap
//-----------------------------------------------------------------------------
void	Cmd_WSrv_Login_f (gentity_t *ent)
{
    int     i;
    char    szMsg[64] ;
	char	name[MAX_INFO_VALUE], password[MAX_INFO_VALUE];
	int		clientNum;

	if ( g_WsrvData.Connected )
    {
        if ( ent )
        {
            // a Login when already connected cause a restart to the STARTMAP
            // If a client type wsrv_Login, he quit the game and recreate his own server
			clientNum = ent - g_entities;
			trap_SendServerCommand(clientNum, va("map %s\n", g_WsrvData.szStartMap) );
        }
        else
			trap_SendConsoleCommand(EXEC_APPEND, va("map %s\n", g_WsrvData.szStartMap) );
        return;
    }

    if ( trap_Argc() != 3 ) 
    {
		Com_Printf("Usage: wsrv_login <name> <password>\n");
		return;
	}

	trap_Cvar_VariableStringBuffer( "wsrv_url", g_WsrvData.szUrl, sizeof(g_WsrvData.szUrl) );
    if ( strlen(g_WsrvData.szUrl) == 0 )
    {
		Com_Printf("wsrv_url must be defined\n");
		return;
	}

	trap_Argv( 1, name, sizeof(name) );
	trap_Argv( 2, password, sizeof(password) );

    RPC_Login(&l_RpcData, name, password) ;

    while ( l_RpcData.byRpcStatus == RPC_STATUS_IN_PROGRESS )
    {
        RPC_Sleep(100) ;
        Com_Printf(".");
    }
    Com_Printf("\n");

    if ( l_RpcData.byRpcStatus == RPC_STATUS_OK )
    {
        if ( RPC_Login_OK(&l_RpcData, szMsg, sizeof(szMsg)) )
        {
            Com_Printf("Connected to WorldServer\n");
			g_WsrvData.Connected = CON_CONNECTING;
            trap_Cvar_Set("name", name);		// FIXME : forceset
            trap_Cvar_Set("hostname", name);	// FIXME : forceset

            for ( i=0 ; i<8 ; i++ ) 
                trap_Cvar_Set(va("adr%d", i), "");

            // Now create the game
            RPC_PlayerCreateGame(&l_RpcData) ;
            while ( l_RpcData.byRpcStatus == RPC_STATUS_IN_PROGRESS )
            {
                RPC_Sleep(100) ;
                Com_Printf(".");
            }
            Com_Printf("\n");

            if ( l_RpcData.byRpcStatus == RPC_STATUS_OK )
            {
                if ( RPC_PlayerCreateGame_OK(&l_RpcData, szMsg, sizeof(szMsg)) )
                {
					trap_SendConsoleCommand(EXEC_APPEND, va("map %s\n", g_WsrvData.szStartMap) );
                }
            }
        }
        else
            Com_Printf("Unable to Login to World Server : %s\n", szMsg);
    }
    else if ( l_RpcData.byRpcStatus == RPC_STATUS_ERROR )
        Com_Printf("Unable to Login to World Server : No reply\n");

    l_RpcData.byRpcStatus = RPC_STATUS_READY;
}

//-----------------------------------------------------------------------------
// Console command : Logout from the WorldServer
// The "ent" parameter could be NULL if it's a dedicated server
//-----------------------------------------------------------------------------
void	Cmd_WSrv_Logout_f(gentity_t *ent)
{
    if ( ent && ent != g_entities )    // Prevent a Logout from a client
		return;

    if ( trap_Argc() ) 
    {
		Com_Printf("Usage: wsrv_logout\n");
		return;
	}

    if ( strlen(g_WsrvData.szUrl) == 0 )
    {
		Com_Printf("wsrv_url must be defined\n");
		return;
	}

    if ( !g_WsrvData.Connected)
    {
		Com_Printf("not connected to WorldServer\n");
		return;
	}

    RPC_Logout(&l_RpcData) ;

    while ( l_RpcData.byRpcStatus == RPC_STATUS_IN_PROGRESS )
    {
        RPC_Sleep(100) ;
        Com_Printf(".");
    }
    Com_Printf("\n");

    if ( l_RpcData.byRpcStatus == RPC_STATUS_OK )
    {
        Com_Printf("Disconnected from WorldServer\n");
        g_WsrvData.Connected = CON_DISCONNECTED;
		trap_Cvar_Set("tag_worldserver", "");
    }
    else if ( l_RpcData.byRpcStatus == RPC_STATUS_ERROR )
        Com_Printf("Unable to Logout from World Server : No reply\n");

    l_RpcData.byRpcStatus = RPC_STATUS_READY;

	trap_SendConsoleCommand(EXEC_APPEND, "disconnect");
}

// ExitLevel : save data of all clients (not required but safe)
void trap_WSRV_ExitLevel()
{
    char    szCmd[32];
	int		i;

	if ( !g_WsrvData.Connected )
		return;

	for ( i=0 ; i<g_maxclients.integer ; i++ )
	{
		if ( level.clients[i].pers.connected )
		{
            Com_sprintf(szCmd, sizeof(szCmd), "SetInventory,%s", level.clients[i].pers.netname);
            WSrv_AddCmd(szCmd);
        }
    }

	// Unlock the map command n time for manage the '+' sign in changemap
//	i = 1;
//	for ( pcChar = level.changemap ; *pcChar ; pcChar++ )
//		if ( *pcChar == '+' )
//			i++ ;
//	WSrv_SetMapLock(i);
}


// ShutdownGame : save data of all clients
void trap_WSRV_ShutdownGame()
{
    char    szCmd[32];
	int		i;
	char	tag_worldserver[1+sizeof(g_WsrvData)*2];

//<worldserver> FIXME : bad hack for preserve worldserver data throught a dll reload
	WSrv_SerializeWsrvData(&g_WsrvData, tag_worldserver, sizeof(tag_worldserver));
	trap_Cvar_Set("tag_worldserver",  tag_worldserver);

	if ( g_WsrvData.Connected != CON_CONNECTED )
		return;

	for ( i=0 ; i<g_maxclients.integer ; i++ )
	{
		if ( level.clients[i].pers.connected )
		{
            Com_sprintf(szCmd, sizeof(szCmd), "SetInventory,%s", level.clients[i].pers.netname);
            WSrv_AddCmd(szCmd);
        }
    }

    WSrv_AddCmd("Logout");
    WSrv_CheckWorldServer(qtrue);    // Flush the command buffer
}

void trap_WSRV_InitGame()
{
	char	tag_worldserver[1+sizeof(g_WsrvData)*2];

	if ( g_WsrvData.Connected == CON_CONNECTED )
		return;

	// FIXME : bad hack for preserve worldserver data throught a dll reload
	trap_Cvar_VariableStringBuffer( "tag_worldserver", tag_worldserver, sizeof(tag_worldserver) );
	if ( strlen(tag_worldserver) )
	{
		WSrv_UnserializeWsrvData(&g_WsrvData, tag_worldserver, sizeof(tag_worldserver));
		g_WsrvData.Connected = CON_CONNECTED;
	}
	else
	    Com_Printf("tag_worldserver not found\n");
}

void trap_WSRV_ClientConnect( int clientNum, qboolean firstTime, qboolean isBot ) 
{
	if ( !g_WsrvData.Connected )
		return;

	if ( firstTime && !isBot )
		WSrv_ClientConnect( &g_entities[ clientNum ] );
}

void trap_WSRV_die( gentity_t *targ, gentity_t *inflictor, gentity_t *attacker, int damage, int meansOfDeath ) 
{
	if ( !g_WsrvData.Connected )
		return;

	WSrv_Killed(targ, inflictor, attacker, meansOfDeath);
}

void trap_WSRV_RunFrame() 
{
	if ( !g_WsrvData.Connected )
		return;

	WSrv_CheckWorldServer(qfalse);
}

//-----------------------------------------------------------------------------
// Access to the worldserver internal data
//-----------------------------------------------------------------------------
const char	*WSrv_GetStartmap()				{ return g_WsrvData.szStartMap; }

const int	WSrv_GetMapLock()				{ return g_WsrvData.iMapLock; }
void		WSrv_SetMapLock(int i_iMapLock)	{ g_WsrvData.iMapLock = i_iMapLock; }

const DWORD	WSrv_GetFlags()					{ return g_WsrvData.dwFlags; }

void		WSrv_SerializeWsrvData(void *pVoid, char *string, int size)	
{
	char	*pChar;
	int		i;

	pChar = (char*)pVoid ;
	for ( i=0 ; i<(size-1)/2 ; i++ )
	{
		Com_sprintf(string, 3, "%02X", *pChar);
		pChar++;
		string += 2;
	}
}

void		WSrv_UnserializeWsrvData(void *pVoid, char *string, int size)
{
	char	*pChar;
	int		i, iValue;

	pChar = (char*)pVoid ;
	for ( i=0 ; i<(size-1)/2 ; i++ )
	{
		sscanf(string, "%02X", &iValue);
		*pChar = (char)iValue;
		pChar++;
		string += 2;
	}
}
 